import math
import random
import sys
import timeout_decorator
sys.path.append('/media/angelo/WorkData/git/VITAMIN-APP/VITAMIN/')
from vitamin_model_checker.model_checker_interface.explicit import ATL
from vitamin_model_checker.utils.generators import CGS_generator
import time

TIMEOUT = 2700  # seconds

# Wrap the functions that require a timeout with the decorator
@timeout_decorator.timeout(TIMEOUT)
def generate_model(agents, states):
    return CGS_generator.generate_random_CGS(agents, states, states, random.randint(1, states), 'tmp')

@timeout_decorator.timeout(TIMEOUT)
def model_checking(phi):
    return ATL.model_checking(phi, 'tmp')

def run_experiments(num_agents_min, num_agents_max, num_states_min, num_states_max, repetitions):
    with open('results.csv', 'w') as file:
        file.write('# States; # Agents; ATL time [sec]\n')    
    for state in range(num_states_min, num_states_max+1, 5):
        for agents in range(num_agents_min, num_agents_max+1):
            coalition = ','.join([str(i) for i in range(1, agents+1)])
            # target = f'a1{random.randint(1, int(math.sqrt(state)))}_{random.randint(1, int(math.sqrt(state)))}'
            target = 'a11_1'
            phi = f'(<{coalition}>F {target})'
            avg_time = 0
            with open('results.csv', 'a') as file:
                file.write(str(state) + ';' + str(agents) + ';')
            for _ in range(0, repetitions):
                try:
                    # Measure start time
                    start = time.time()
                    # Generate the model
                    generate_model(agents, int(math.sqrt(state)))
                    # Run model checking
                    res = model_checking(phi)
                    # Measure end time
                    end = time.time()
                    avg_time += end - start
                except timeout_decorator.TimeoutError:
                    print("Function took too long and was terminated.")
                    avg_time += TIMEOUT  # Consider the timeout duration as the time taken for this run
            avg_time = avg_time / repetitions
            with open('results.csv', 'a') as file:
                file.write(str(avg_time) + '\n')

if __name__ == "__main__":
    num_agents_min = max(int(sys.argv[1]), 1)
    num_agents_max = max(int(sys.argv[2]), 1)
    num_states_min = max(int(sys.argv[3]), 1)
    num_states_max = max(int(sys.argv[4]), 1)
    repetitions = max(int(sys.argv[5]), 1)
    run_experiments(num_agents_min, num_agents_max, num_states_min, num_states_max, repetitions)



import itertools
import random
from vitamin_model_checker.models.CGS import *
import os
from vitamin_model_checker.model_checker_interface.explicit.NatATL.NatATLtoCTL import *

found_solution = False

#this function returns one strategy at time using "yield" keyword
def generate_conditions(P, C, max_k):
    condition_set = set()

    def generate_condition(k, condition):
        if k == 0:
            condition_str = ' && '.join(condition)
            if condition_str not in condition_set:
                yield condition_str
                condition_set.add(condition_str)
        else:
            for p in P:
                if p not in condition:
                    new_condition = condition + [p]
                    if len(new_condition) == 1:
                        yield from generate_condition(k - 1, new_condition)
                    elif len(new_condition) > 1:
                        new_condition.sort()
                        new_condition_str = new_condition[0] + " " + random.choice(C) + " "
                        for i in range(1, len(new_condition) - 1):
                            new_condition_str += new_condition[i] + " " + random.choice(C) + " "
                        new_condition_str += new_condition[-1]
                        complexity = len(new_condition_str.split())
                        if complexity <= max_k:
                            yield from generate_condition(k - 1, [new_condition_str])

    for k in range(1, max_k + 1):
        yield from generate_condition(k, [])

def generate_negated(input_list, max_k):
    for input_str in input_list:
        atomic_props = input_str.split(' && ')
        for combo in itertools.product(['', '!'], repeat=len(atomic_props)):
            negated_props = [f'{combo[i]}{atomic_props[i]}' for i in range(len(atomic_props))]
            new_str = ' && '.join(negated_props)
            complexity = len(new_str.split())
            if "!" in new_str:
                complexity += 1
            if complexity <= max_k:
                yield new_str


def generate_strategies(cartesian_products, k, agents, found_solution):
    strategies = [list() for _ in range(len(agents))]  # Le strategie sono liste

    def search_solution(strategies, current_strategy, depth):
        if depth == len(agents):
            yield current_strategy
        else:
            for agent in strategies[depth]:
                current_strategy.append(agent)
                yield from search_solution(strategies, current_strategy, depth + 1)
                current_strategy.pop()

    if not found_solution:
        for index, agent_key in enumerate(cartesian_products):
            cartesian_product = cartesian_products[agent_key]

            # Generate combinations of all conditions length to reach the desired complexity bound
            for r in range(1, k + 1):
                combinations = itertools.combinations(cartesian_product, r)
                filtered_combinations = [combination for combination in combinations
                                         if len(set(action for _, action in combination)) == r  #equal to r ensures the number of different actions to be reached
                                         ]
                for combination in filtered_combinations:
                    total_complexity = sum(
                        len(condition.split()) + (1 if "!" in condition else 0) for condition, _ in combination)
                    if total_complexity == k:
                        new_strategy = {"condition_action_pairs": list(combination)}
                        if not is_duplicate(strategies[index], new_strategy):
                            strategies[index].append(new_strategy)
                            yield from search_solution(strategies, [], 0)

    return strategies


def is_duplicate(existing_dictionaries, new_dictionary):
    for existing_dictionary in existing_dictionaries:
        if existing_dictionary['condition_action_pairs'] == new_dictionary['condition_action_pairs']:
            return True
    return False


class BreakLoop(Exception):
    pass

def agent_combinations(new_combinations):
    for agent1 in new_combinations:
        for agent2 in new_combinations:
                yield agent1, agent2

def initialize(model_path, formula):
    filename = os.path.abspath(model_path)
    if not os.path.isfile(filename):
        raise FileNotFoundError(f"No such file or directory: {filename}")
    cgs = CGS()
    cgs.read_file(filename)

    #For testing only: randomize formula REMOVE THE COMMENTS FROM THE CURRENT TO THE NEXT 4 ROWS USE A RANDOMIZED FORMULA AT EACH EXECUTION
    #randomized_formula = randomize_natatl_formula(formula, get_number_of_agents(), get_atomic_prop())
    #with open(f'C:\\Users\\utente\\Desktop\\Tesi\\TESTING\\Exists next with n agents\\testing11\\modified_formula.txt','w') as f:
     #   f.write(str(randomized_formula))
    #print(randomized_formula)

    #check if input model is correct
    cgs.matrixParser(cgs.get_number_of_agents())
    #transform natATL formula into CTL formula
    CTLformula = natatl_to_ctl(formula)
    print(formula)
    print(CTLformula)
    k = get_k_value(formula)

    agents = get_agents_from_natatl(formula)
    print(f"Envolved agents: {agents}")
    actions_per_agent = cgs.get_actions(agents)
    print(f"actions picked by each agent:{actions_per_agent}")
    agent_actions = {}
    for i, agent_key in enumerate(actions_per_agent.keys()):
        agent_actions[f"actions_{agent_key}"] = actions_per_agent[agent_key]
    #print(f"Obtained actions: {agent_actions}")
    actions_list = [actions for actions in agent_actions.values()]
    atomic_propositions = cgs.get_atomic_prop()
    print(atomic_propositions)
    return k, agent_actions, actions_list, atomic_propositions, CTLformula, agents, cgs

def generate_cartesian_products(actions_list, conditions):
    cartesian_products = {}
    for i, actions in enumerate(actions_list, start=1):
        agent_key = f"actions_agent{i}"
        agent_cartesian_product = list(itertools.product(conditions, actions))
        if agent_key not in cartesian_products:
            cartesian_products[agent_key] = []
        cartesian_products[agent_key].extend(agent_cartesian_product)
    return cartesian_products

def generate_guarded_action_pairs(k, agent_actions, actions_list, atomic_propositions):
    C = ['and', 'or']
    try:
        cartesian_products = {}
        for agent_key in agent_actions.keys():
            conditions = list(generate_conditions(atomic_propositions, C, k))
            for condition in conditions:
                negated_conditions = list(generate_negated([condition], k))
                all_conditions = [condition] + negated_conditions
                new_cartesian_products = generate_cartesian_products(actions_list, all_conditions)
                for key, value in new_cartesian_products.items():
                    if key not in cartesian_products:
                        cartesian_products[key] = []
                    cartesian_products[key].extend(value)
        return cartesian_products

    except Exception as e:
        print(f"An error occurred: {e}")
        return {}






# parsetab.py
# This file is automatically generated. Do not edit.
# pylint: disable=W,C,R
_tabversion = '3.10'

_lr_method = 'LALR'

_lr_signature = 'AND COALITION_DEMONIC EVENTUALLY FALSE GLOBALLY IMPLIES LPAREN NEXT NOT OR PROP RELEASE RPAREN TRUE UNTIL WEAKexpression : expression AND expression\n                  | expression OR expression\n                  | expression IMPLIES expressionexpression : COALITION_DEMONIC expression UNTIL expression\n                  | COALITION_DEMONIC expression WEAK expression\n                  | COALITION_DEMONIC expression RELEASE expressionexpression : COALITION_DEMONIC GLOBALLY expression\n                  | COALITION_DEMONIC NEXT expression\n                  | COALITION_DEMONIC EVENTUALLY expressionexpression : NOT expressionexpression : LPAREN expression RPARENexpression : FALSE \n                  | TRUEexpression : PROP'
    
_lr_action_items = {'COALITION_DEMONIC':([0,2,3,4,8,9,10,12,13,14,20,21,22,],[2,2,2,2,2,2,2,2,2,2,2,2,2,]),'NOT':([0,2,3,4,8,9,10,12,13,14,20,21,22,],[3,3,3,3,3,3,3,3,3,3,3,3,3,]),'LPAREN':([0,2,3,4,8,9,10,12,13,14,20,21,22,],[4,4,4,4,4,4,4,4,4,4,4,4,4,]),'FALSE':([0,2,3,4,8,9,10,12,13,14,20,21,22,],[5,5,5,5,5,5,5,5,5,5,5,5,5,]),'TRUE':([0,2,3,4,8,9,10,12,13,14,20,21,22,],[6,6,6,6,6,6,6,6,6,6,6,6,6,]),'PROP':([0,2,3,4,8,9,10,12,13,14,20,21,22,],[7,7,7,7,7,7,7,7,7,7,7,7,7,]),'$end':([1,5,6,7,15,17,18,19,23,24,25,26,27,28,29,],[0,-12,-13,-14,-10,-1,-2,-3,-7,-8,-9,-11,-4,-5,-6,]),'AND':([1,5,6,7,11,15,16,17,18,19,23,24,25,26,27,28,29,],[8,-12,-13,-14,8,8,8,8,8,8,8,8,8,-11,8,8,8,]),'OR':([1,5,6,7,11,15,16,17,18,19,23,24,25,26,27,28,29,],[9,-12,-13,-14,9,9,9,9,9,9,9,9,9,-11,9,9,9,]),'IMPLIES':([1,5,6,7,11,15,16,17,18,19,23,24,25,26,27,28,29,],[10,-12,-13,-14,10,10,10,10,10,10,10,10,10,-11,10,10,10,]),'GLOBALLY':([2,],[12,]),'NEXT':([2,],[13,]),'EVENTUALLY':([2,],[14,]),'UNTIL':([5,6,7,11,15,17,18,19,23,24,25,26,27,28,29,],[-12,-13,-14,20,-10,-1,-2,-3,-7,-8,-9,-11,-4,-5,-6,]),'WEAK':([5,6,7,11,15,17,18,19,23,24,25,26,27,28,29,],[-12,-13,-14,21,-10,-1,-2,-3,-7,-8,-9,-11,-4,-5,-6,]),'RELEASE':([5,6,7,11,15,17,18,19,23,24,25,26,27,28,29,],[-12,-13,-14,22,-10,-1,-2,-3,-7,-8,-9,-11,-4,-5,-6,]),'RPAREN':([5,6,7,15,16,17,18,19,23,24,25,26,27,28,29,],[-12,-13,-14,-10,26,-1,-2,-3,-7,-8,-9,-11,-4,-5,-6,]),}

_lr_action = {}
for _k, _v in _lr_action_items.items():
   for _x,_y in zip(_v[0],_v[1]):
      if not _x in _lr_action:  _lr_action[_x] = {}
      _lr_action[_x][_k] = _y
del _lr_action_items

_lr_goto_items = {'expression':([0,2,3,4,8,9,10,12,13,14,20,21,22,],[1,11,15,16,17,18,19,23,24,25,27,28,29,]),}

_lr_goto = {}
for _k, _v in _lr_goto_items.items():
   for _x, _y in zip(_v[0], _v[1]):
       if not _x in _lr_goto: _lr_goto[_x] = {}
       _lr_goto[_x][_k] = _y
del _lr_goto_items
_lr_productions = [
  ("S' -> expression","S'",1,None,None,None),
  ('expression -> expression AND expression','expression',3,'p_expression_binary','parser.py',55),
  ('expression -> expression OR expression','expression',3,'p_expression_binary','parser.py',56),
  ('expression -> expression IMPLIES expression','expression',3,'p_expression_binary','parser.py',57),
  ('expression -> COALITION_DEMONIC expression UNTIL expression','expression',4,'p_expression_ternary','parser.py',66),
  ('expression -> COALITION_DEMONIC expression WEAK expression','expression',4,'p_expression_ternary','parser.py',67),
  ('expression -> COALITION_DEMONIC expression RELEASE expression','expression',4,'p_expression_ternary','parser.py',68),
  ('expression -> COALITION_DEMONIC GLOBALLY expression','expression',3,'p_expression_unary','parser.py',73),
  ('expression -> COALITION_DEMONIC NEXT expression','expression',3,'p_expression_unary','parser.py',74),
  ('expression -> COALITION_DEMONIC EVENTUALLY expression','expression',3,'p_expression_unary','parser.py',75),
  ('expression -> NOT expression','expression',2,'p_expression_not','parser.py',80),
  ('expression -> LPAREN expression RPAREN','expression',3,'p_expression_group','parser.py',85),
  ('expression -> FALSE','expression',1,'p_expression_boolean','parser.py',89),
  ('expression -> TRUE','expression',1,'p_expression_boolean','parser.py',90),
  ('expression -> PROP','expression',1,'p_expression_prop','parser.py',94),
]



from vitamin_model_checker.model_checker_interface.explicit.CapATL.CapATL import *
from vitamin_model_checker.model_checker_interface.explicit.CapATL.classes import *
from vitamin_model_checker.model_checker_interface.explicit.CapATL.pre import *


class p_knowledge_for_Y:

    def __init__(self, state, knowledge, beta, agents, agents_tot):
        self.state = state
        self.knowledge = knowledge
        self.action = beta 
        self.coalition = agents
        self.agents = agents_tot
    
    def print_p_knowledge_for_Y(self) :
        print(self.state)
        interm = self.agents
        if type(interm) == str:
            interm = self.agents.split(',')
        for i in interm :
            print("Ag" , int(i) , self.knowledge[int(i)-1])
        print(self.action)
        return ""

    def not_in(self,ens) :
        for elem in ens : 
            if getattr(elem, 'state') == self.state and getattr(elem, 'knowledge') == self.knowledge and getattr(elem, 'coalition') == self.coalition and getattr(elem, 'action') == self.action:
                return False 
        return True



class p_knowledge:
    def __init__(self, state, knowledge, agents):
        self.state = state
        self.knowledge = knowledge
        self.agents = agents
        
    def print_p_knowledge(self):
        print(self.state, self.knowledge, self.agents)
    
     # returns true if self is not in ens
    def not_in(self, ens) :
        for elem in ens :
            if getattr(elem, 'state') == self.state and getattr(elem, 'knowledge') == self.knowledge :
                return False 
        return True  
    
    def __eq__(self, other):
        if isinstance(other, p_knowledge):
            return self.state == other.state and self.knowledge == other.knowledge and self.agents == other.agents
        return False

    def __hash__(self):
        return hash((self.state, self.knowledge, self.agents))


class Node_PK:
   def __init__(self, data):
      self.left = None
      self.right = None
      self.value = data



class Parser:
    def __init__(self, input):
        self.tokens = input.split()
        self.current_token = None
        self.next()

    def next(self):
        if len(self.tokens) == 0:
            self.current_token = None
        else:
            self.current_token = self.tokens.pop(0)

    def parse_expression(self):
        result = self.parse_implies_expression()
        while self.current_token == 'iff':
            self.next()
            result = f"({result} iff {self.parse_implies_expression()})"
        return result

    def parse_implies_expression(self):
        result = self.parse_or_expression()
        while self.current_token == 'implies':
            self.next()
            result = f"({result} implies {self.parse_or_expression()})"
        return result

    def parse_or_expression(self):
        result = self.parse_and_expression()
        while self.current_token == 'or':
            self.next()
            result = f"({result} or {self.parse_and_expression()})"
        return result

    def parse_and_expression(self):
        result = self.parse_unary_expression()
        while self.current_token == 'and':
            self.next()
            result = f"({result} and {self.parse_unary_expression()})"
        return result

    def parse_unary_expression(self):
        if self.current_token == 'not':
            self.next()
            return f"(not {self.parse_term()})"
        else:
            return self.parse_term()

    def parse_term(self):
        term = self.current_token
        self.next()
        return term

def parse_string(input_str):
    parser = Parser(input_str)
    return parser.parse_expression()


import xml.etree.ElementTree as ET

#The transition in mulval are calculated from teh end to the beginning, we then sweep the src and dst of each transition
class Transition:
    def __init__(self, src, dst):
        self.src = dst
        self.dst = src

# A node has an id, a fact (the knowledge or attack step it represent) and a type (if it's an attack or not). It may have a metric              
class Node: 
    def __init__(self, id, fact, metric, type):
        self.id = id
        self.fact = fact
        self.metric = metric
        self.type = type
        
#from the xml file we get the list of nodes and of transitions  
def build_data(file):
    tree = ET.parse(file)
    root = tree.getroot()
    transition = []
    vertices = []

    #we get the transitions
    for arc in root[0].findall('arc'):
        src = int(arc.find('src').text)
        dst = int(arc.find('dst').text)
        transition.append(Transition(src, dst))
    
    #we get the nodes
    for vertex in root[1].findall('vertex'):
        id = int(vertex.find('id').text) - 1
        fact = vertex.find('fact').text
        metric = int(vertex.find('metric').text)
        type = vertex.find('type').text
        vertices.append(Node(id, fact, metric, type))
        
    return (transition, vertices)





import time

from vitamin_model_checker.model_checker_interface.explicit import RBATL
start = time.time()
result = RBATL.model_checking('<1><0,0>F(oc && rm && <1><0,0>F((pl || pr) && <1><0,0>F(oc && rm)))', './examples/RBATL/rover.txt')
end = time.time()
print('phi1')
print(result)
print('TIME:', end-start, 'seconds')
# start = time.time()
# result = RBATL.model_checking('<1,2><5,1>F(oc && rm && <1><0,0>F((pl || pr) && <1><0,0>F(oc && rm)))', './examples/RBATL/rover.txt')
# end = time.time()
# print('phi2')
# print(result)
# print('TIME:', end-start, 'seconds')
# start = time.time()
# result = RBATL.model_checking('<1,2><2,2>F(oc && rm && <1><0,0>F((pl || pr) && <1><0,0>F(oc && rm)))', './examples/RBATL/rover.txt')
# end = time.time()
# print('phi2_rb')
# print(result)
# print('TIME:', end-start, 'seconds')
# del RBATL

# from vitamin_model_checker.model_checker_interface.explicit.RABATL import RABATL
# print(RABATL.model_checking('<1,2,3,4><4>F(p1 && q2 && s4 && r3)', 'tmp'))
# <1,2,3,4><10>F(p0 && q1 && q2 && r2 && p2 && p3)





# for i in range(2, 11):
#     print('N. of resources: ', i)
#     start = time.time()
#     result  = RABATL.model_checking('(<1,2><{res}>F p)'.format(res=','.join(['2']*i)), f'./examples/RABATL/sensor_network{i}.txt')
#     end = time.time()
#     # print(result)
#     print('TIME:', end-start, 'seconds')




# RB-ATL
# Resources: 1 energy, 1 memory 
# TIME: 0.006029605865478516 seconds
# Resources: 2 energy, 2 memory 
# TIME: 0.21443557739257812 seconds
# Resources: 3 energy, 3 memory 
# TIME: 7.137519121170044 seconds
# Resources: 4 energy, 4 memory 
# TIME: 261.43060302734375 seconds

# RAB-ATL
# Resources: 1 energy, 1 memory 
# TIME: 0.007246255874633789 seconds
# Resources: 2 energy, 2 memory 
# TIME: 0.19492864608764648 seconds
# Resources: 3 energy, 3 memory 
# TIME: 5.23409366607666 seconds
# Resources: 4 energy, 4 memory 
# TIME: 177.9505763053894 seconds



# start = time.time()
# result  = RBATL.model_checking('(<1,2><2,2,2,2,2,2,2,2,2,2>F p)', './examples/RBATL/sensor_network4.txt')
# end = time.time()
# print(result)
# print('TIME:', end-start, 'seconds')
# result  = RBATL.model_checking('(<1,2,3><2,2>F h)', './examples/RBATL/RBATL_model.txt')

# n = 10

# for energy in range(1, n):
#     # for memory in range(1, n):
#     print('Resource bound', energy)
#     # print('Memory resource bound', memory)
#     phi = f'(<1,2><{energy},{energy}>F p)'
#     start = time.time()
#     result  = RBATL.model_checking(phi, './examples/RBATL/sensor_network.txt')
#     print('RBATL')
#     # print(result)
#     print(f'TIME: {time.time()-start} seconds')
    
#     start = time.time()
#     result  = RABATL.model_checking(phi, './examples/RABATL/sensor_network.txt')
#     print('RABATL')
#     # print(result)
#     print(f'TIME: {time.time()-start} seconds')

    # result  = RABATL.model_checking(phi, './examples/RABATL/sensor_network_mod.txt')
    # print('RABATL mod')
    # print(result)
    # print(f'TIME: {time.time()-start} seconds')


import random
import sys

def generate_transition_matrix(num_states):
    transitions = []
    for _ in range(num_states):
        row = []
        for _ in range(num_states):
            # if random.random() < 0.3:  # Adjust the probability as needed
            #     row.append('0')
            # else:
            row.append(''.join(random.choices(['A', '^', 'B'], k=2)))
        transitions.append(row)
    return transitions

def generate_name_states(num_states):
    return ['s' + str(i) for i in range(num_states)]

def generate_initial_state(states):
    return random.choice(states)

def generate_costs_for_actions(states, transition_matrix, num_resources, num_agents):
    costs = []
    cost_dict = {}
    i = 0
    for state in states:
        for action in transition_matrix[i]:
            if action != '0':
                if action not in cost_dict:
                    cost_dict[action] = {}
                resources = []
                for _ in range(0, num_resources):
                    resources.append(','.join([str(random.randint(1, 5)) for _ in range(num_agents)]))
                cost_dict[action][state] = '$' + ':'.join(resources)
        i += 1    
    for action in cost_dict:
        costs.append(action + ' ')
        cost_str = []
        for state in cost_dict[action]:
            cost_str.append(state + cost_dict[action][state])
        costs.append(';'.join(cost_str))
        costs.append('\n')
    return costs

def generate_atomic_propositions():
    return ['p', 'q']

def generate_labelling(num_states, num_atomic_propositions):
    return [[random.randint(0, 1) for _ in range(num_atomic_propositions)] for _ in range(num_states)]

def generate_number_of_agents():
    return 2
    # return random.randint(1, 10)  # Adjust the range as needed

def generate_random_model(num_states, num_resources):
    num_agents = generate_number_of_agents()
    transition_matrix = generate_transition_matrix(num_states)
    name_states = generate_name_states(num_states)
    initial_state = generate_initial_state(name_states)
    costs_for_actions = generate_costs_for_actions(name_states, transition_matrix, num_resources, num_agents)
    atomic_propositions = generate_atomic_propositions()
    labelling = generate_labelling(num_states, len(atomic_propositions))

    model = "Transition\n{tr}".format(tr=''.join([' '.join(row) + '\n' for row in transition_matrix]))
    model += "Name_State\n{st}\n".format(st=' '.join(name_states))
    model += f"Initial_State\n{initial_state}\n"
    model += "Costs_for_actions_split\n{ct}".format(ct=''.join(costs_for_actions))
    model += f"Atomic_propositions\n{' '.join(atomic_propositions)}\n"
    model += "Labelling\n{lb}".format(lb=''.join([' '.join(map(str, row)) + '\n' for row in labelling]))
    model += f"Number_of_agents\n{num_agents}\n"

    return model

def generate_random_model_file(num_states, num_resources, filename):
    model = generate_random_model(num_states, num_resources)
    with open(filename, 'w') as f:
        f.write(model)

if __name__ == "__main__":
    num_states = max(int(sys.argv[1]), 1)
    num_resources = max(int(sys.argv[2]), 1)
    filename = sys.argv[3]
    generate_random_model_file(num_states, num_resources, filename)
    print(f"Random model with {num_states} states and {num_resources} resources generated and saved to {filename}.")



from vitamin_model_checker.model_checker_interface.explicit.NatATL.strategies import initialize, generate_strategies, generate_guarded_action_pairs
from vitamin_model_checker.model_checker_interface.explicit.NatATL.pruning import pruning

def model_checking(formula, model):
    # Start timer
    # start_time = time.time()
    found_solution = False
    result = {}
    # initializes conditions and actions for involved agents
    k, agent_actions, actions_list, atomic_propositions, CTLformula, agents, cgs = initialize(model, formula)
    i = 1
    while not found_solution and i <= k:
        #generate guarded action pairs for each agent
        cartesian_products = generate_guarded_action_pairs(i, agent_actions, actions_list, atomic_propositions)
        # generates the initial strategies
        strategies_generator = generate_strategies(cartesian_products, i, agents, found_solution)
        # check for each strategy if there's a solution via pruning & model checking
        for current_strategy in strategies_generator:  # Iterate over the generator object
            found_solution = pruning(cgs, model, agents, CTLformula, current_strategy)  # Call pruning here
            if found_solution:
                print("Solution found!")
                found_solution = True

                # Write to file
                # with open(result, 'a') as f: #'a' to append results in the same file, 'w' to overwrite the old ones
                result['Satisfiability'] = found_solution
                result['Complexity Bound'] = i
                result['Winning Strategy per agent'] = current_strategy
                print(f"Winning Strategy per agent: {current_strategy}")
                break
        i += 1
    else:
        if (not found_solution):
            print(f"False, no states satisfying {CTLformula} have been found!")
            # with open(result, 'a') as f:
            result['Satisfiability'] = found_solution
            result['Complexity Bound'] = k

    # End timer
    # end_time = time.time()
    # elapsed_time = end_time - start_time
    # print(f"Elapsed time is {elapsed_time} seconds.")

    # # Write execution time to file
    # with open(result, 'a') as f:
    #     f.write(f"Execution time: {elapsed_time} seconds\n")
    
    return result




from vitamin_model_checker.model_checker_interface.explicit.NatATL.NatATL import *
from vitamin_model_checker.model_checker_interface.explicit.NatATL.pruning import *
from vitamin_model_checker.model_checker_interface.explicit.NatATL.strategies import *
from vitamin_model_checker.model_checker_interface.explicit.NatATL.NatATLtoCTL import *


import re
import ply.lex as lex
import ply.yacc as yacc

MAX_COALITION = 0

# Token
tokens = (
    'LPAREN',
    'RPAREN',
    'AND',
    'OR',
    'NOT',
    'IMPLIES',
    'UNTIL',
    'GLOBALLY',
    'NEXT',
    'EVENTUALLY',
    'PROP',
    'COALITION'
)


# Regular expressions for tokens
t_LPAREN = r'\('
t_RPAREN = r'\)'
t_AND = r'&&|\&|and'
t_OR = r'\|\||\||or'
t_NOT = r'!|not'
t_IMPLIES = r'->|>|implies'
t_PROP = r'[a-z]+'
t_UNTIL = r'U|until'
t_GLOBALLY = r'G|globally|always'
t_NEXT = r'X|next'
t_EVENTUALLY = r'F|eventually'
t_COALITION = r'<\{((?:\d+,)*\d+)\},\s*(\d+)>'


# Token error handling
def t_error(t):
    t.lexer.skip(1)



# Grammar
def p_expression_binary(p):
    '''expression : expression AND expression
                  | expression OR expression
                  | expression IMPLIES expression'''
    p[0] = (p[2], p[1], p[3])


class CoalitionValueError(Exception):
    pass

def p_expression_ternary(p):
    '''expression : COALITION expression UNTIL expression'''
    matches = re.findall(r'\{((?:\d+,)*\d+)\},\s*(\d+)', p[1])
    if matches:
        coalition_values, k_value = matches[0]
        for value in coalition_values.split(','):
            if int(value) > int(MAX_COALITION):
                raise CoalitionValueError(f"Coalition value {value} exceeds maximum of {MAX_COALITION}")
        p[0] = (f"<{{{coalition_values}}}, {k_value}>{p[3]}", p[2], p[4])

def p_expression_unary(p):
    '''expression : COALITION GLOBALLY expression
                  | COALITION NEXT expression
                  | COALITION EVENTUALLY expression'''
    matches = re.findall(r'\{((?:\d+,)*\d+)\},\s*(\d+)', p[1])
    if matches:
        coalition_values, k_value = matches[0]
        for value in coalition_values.split(','):
            if int(value) > int(MAX_COALITION):
                raise CoalitionValueError(f"Coalition value {value} exceeds maximum of {MAX_COALITION}")
        p[0] = (f"<{{{coalition_values}}}, {k_value}>{p[2]}", p[3])


def p_expression_not(p):
    '''expression : NOT expression'''
    p[0] = (p[1], p[2])


def p_expression_group(p):
    '''expression : LPAREN expression RPAREN'''
    p[0] = p[2]


def p_expression_prop(p):
    '''expression : PROP'''
    p[0] = p[1]


def p_error(p):
    pass


# lexer and parser
lexer = lex.lex()
parser = yacc.yacc()


def get_lexer():
    return lexer


# given a natATL formula as input and the max number of agents in the model,
# returns a tuple representing the formula divided into subformulas.
def do_parsing(formula, n_agent):
    global MAX_COALITION
    MAX_COALITION = n_agent
    try:
        result = parser.parse(formula)
        print(result)
        return result
    except SyntaxError:  # if parser fails
        return None
    except CoalitionValueError:  # coalition not existent
      return None


# checks whether the input operator corresponds to a given operator defined in the grammar
def verify(token_name, string):
    lexer.input(string)
    for token in lexer:
        if token.type == token_name and token.value in string:
            return True
    return False

#use example: "formula = "<{1,2}, 5> (a && b) U (c || !d)",
# scritta anche come "('<{1,2},5>U', ('&&', 'a', 'b'), ('||', 'c', ('!', 'd')))"


from binarytree import Node
from vitamin_model_checker.models.CGS.CGS import *
from vitamin_model_checker.logics.ATL import *


# returns the states where the proposition holds
def get_states_prop_holds(prop):
    states = set()
    prop_matrix = cgs.get_matrix_proposition()

    index = cgs.get_atom_index(prop)
    if index is None:
        return None
    for state, source in enumerate(prop_matrix):
        if source[int(index)] == 1:
            states.add(state)
    return states



# set of states (ex. s1, s2) as input and returns a set of indices to identify them
def convert_state_set(state_set):
    states = set()
    for elem in state_set:
        position = cgs.get_index_by_state_name(elem)
        states.add(int(position))
    return states


# converts a string into a set
def string_to_set(string):
    if string == 'set()':
        return set()
    set_list = string.strip("{}").split(", ")
    new_string = "{" + ", ".join(set_list) + "}"
    return eval(new_string)


#  function that builds a formula tree, used by the model checker
def build_tree(tpl):
    if isinstance(tpl, tuple):
        root = Node(tpl[0])
        if len(tpl) > 1:
            left_child = build_tree(tpl[1])
            if left_child is None:
                return None
            root.left = left_child
            if len(tpl) > 2:
                right_child = build_tree(tpl[2])
                if right_child is None:
                    return None
                root.right = right_child
    else:
        states = set()
        states_proposition = get_states_prop_holds(str(tpl))
        if states_proposition is None:
            return None
        else:
            for element in states_proposition:
                states.add(cgs.get_state_name_by_index(element))
            root = Node(str(states))
    return root


# It returns the states from which the coalition has a strategy to enforce the next state to lie in state_set.
# function used by the model checker.
def pre(coalition, state_set):
    agents = cgs.get_agents_from_coalition(coalition)
    graph = cgs.get_graph()
    state_set = convert_state_set(state_set)  # returns a set of indexes
    pre_states = set()
    dict_state_action = dict()  # dictionary state-action
    for i, source in enumerate(graph):  # take states that have at least one transition to one of the states in the set
        for j in state_set:
            if graph[i][j] != 0:
                coordinates = str(i) + "," + str(j)
                dict_state_action.update({coordinates: cgs.build_list(graph[i][j])})

    # iterate over these states and check that my move is not present in any other state
    for key, value in dict_state_action.items():
        other_actions_in_row = dict()  # dictionary containing other moves of the row (excluding those saved in dict_state_action).
        all_actions_in_row = set()  # all elements in the row
        i = int(key.split(',')[0])
        j = int(key.split(',')[1])
        for index, element in enumerate(graph[i]):
            if element != 0:
                coordinates = str(i) + "," + str(index)
                all_actions_in_row.update(cgs.build_list(graph[i][index]))
                if index != j:
                    other_actions_in_row.update({coordinates: cgs.build_list(graph[i][index])})

        # check if there is a loop -> if yes, it is a pre
        for y in value:
            if '*' in y:
                pre_states.add(str(i))
                break

        for action in value:
            move = cgs.get_coalition_action(set([action]), agents)
            # checks whether a move is present in the others
            check_passed = True
            for k, v in other_actions_in_row.items():
                other_act = cgs.get_coalition_action(v, agents)
                if move.intersection(other_act): # they have something in common
                    # check if it belongs to the set
                    column = int(k.split(',')[1])
                    if column not in state_set:
                        check_passed = False
            if check_passed == True:
                pre_states.add(i)

    # take the elements of the set and check if the moves towards the state contain all the possible opponent moves
    result = set()
    for i in pre_states:
        i = int(i)
        opponent_moves_in_state = set()
        opponent_moves_in_row = set()
        for j, column in enumerate(graph[int(i)]):
            if graph[i][j] != 0:
                opponent_moves_in_row.update(cgs.get_opponent_moves(cgs.build_list(graph[i][j]), agents))

                if j in state_set:
                    # takes the opponent's moves for actions towards the states in state_set
                    opponent_moves_in_state.update(cgs.get_opponent_moves(cgs.build_list(graph[i][j]), agents))

        if opponent_moves_in_row == opponent_moves_in_state:
            # convert i into corresponding state
            result.add(cgs.get_state_name_by_index(i))
    return result


# function that solves the formula tree. The result is the model checking result.
# It solves every node depending on the operator.
def solve_tree(node):
    if node.left is not None:
        solve_tree(node.left)
    if node.right is not None:
        solve_tree(node.right)

    if node.right is None:   # UNARY OPERATORS: not, globally, next, eventually
        if verify('NOT', node.value):  # e.g. ¬φ
            states = string_to_set(node.left.value)
            all_states = set(cgs.get_states())
            ris = all_states - states
            node.value = str(ris)

        elif verify('COALITION', node.value) and verify('GLOBALLY', node.value):  # e.g. <1>Gφ
            coalition = node.value[1:-2]
            states = string_to_set(node.left.value)
            p = set(cgs.get_states())
            t = states
            while p - t:  # p not in t
                p = t
                t = pre(coalition, p) & states
            node.value = str(p)

        elif verify('COALITION', node.value) and verify('NEXT', node.value):  # e.g. <1>Xφ
            coalition = node.value[1:-2]
            states = string_to_set(node.left.value)
            ris = pre(coalition, states)
            node.value = str(ris)

        elif verify('COALITION', node.value) and verify('EVENTUALLY', node.value):  # e.g. <1>Fφ
            # trueUϕ.
            coalition = node.value[1:-2]
            states = string_to_set(node.left.value)
            p = set()
            t = states
            while t - p:  # t not in p
                p.update(t)
                t = pre(coalition, p)
            node.value = str(p)

    if node.left is not None and node.right is not None:  # BINARY OPERATORS: or, and, until, implies
        if verify('OR', node.value): # e.g. φ || θ
            states1 = string_to_set(node.left.value)
            states2 = string_to_set(node.right.value)
            ris = states1.union(states2)
            node.value = str(ris)

        elif verify('COALITION', node.value) and verify('UNTIL', node.value):  # e.g. <1>φUθ
            coalition = node.value[1:-2]
            states1 = string_to_set(node.left.value)
            states2 = string_to_set(node.right.value)
            p = set()
            t = states2
            while t - p:  # t not in p
                p.update(t)
                t = pre(coalition, p) & states1
            node.value = str(p)

        elif verify('AND', node.value):  # e.g. φ && θ
            states1 = string_to_set(node.left.value)
            states2 = string_to_set(node.right.value)
            ris = states1.intersection(states2)
            node.value = str(ris)

        elif verify('IMPLIES', node.value):  # e.g. φ -> θ
            # p -> q ≡ ¬p ∨ q
            states1 = string_to_set(node.left.value)
            states2 = string_to_set(node.right.value)
            not_states1 = set(cgs.get_states()).difference(states1)
            ris = not_states1.union(states2)
            node.value = str(ris)

# returns whether the result of model checking is true or false in the initial state
def verify_initial_state(initial_state, string):
    if initial_state in string:
        return True
    return False


# does the parsing of the model, the formula, builds a tree and then it returns the result of model checking
# function called by front_end_CS
def model_checking(formula, filename):
    global cgs

    if not formula.strip():
        result = {'res': 'Error: formula not entered', 'initial_state': ''}
        return result

    # model parsing
    cgs = CGS()
    cgs.read_file(filename)

    # formula parsing
    res_parsing = do_parsing(formula, cgs.get_number_of_agents())
    if res_parsing is None:
        result = {'res': "Syntax Error", 'initial_state': ''}
        return result
    root = build_tree(res_parsing)
    if root is None:
        result = {'res': "Syntax Error: the atom does not exist", 'initial_state': ''}
        return result

    # model checking
    solve_tree(root)

    # solution
    initial_state = cgs.get_initial_state()
    bool_res = verify_initial_state(initial_state, root.value)
    result = {'res': 'Result: ' + str(root.value), 'initial_state': 'Initial state '+ str(initial_state) + ": " + str(bool_res)}
    return result





import re
import ply.lex as lex
import ply.yacc as yacc

# Token
tokens = (
    'LPAREN',
    'RPAREN',
    'AND',
    'OR',
    'NOT',
    'IMPLIES',
    'UNTIL',
    'RELEASE',
    'WEAK',
    'GLOBALLY',
    'NEXT',
    'EVENTUALLY',
    'FALSE',
    'TRUE',
    'PROP',
    'DEMONIC'
)

# Regular expressions for tokens
t_LPAREN = r'\('
t_RPAREN = r'\)'
t_AND = r'&&|\&|and'
t_OR = r'\|\||\||or'
t_NOT = r'!|not'
t_IMPLIES = r'->|>|implies'
t_UNTIL = r'U|until'
t_RELEASE = r'R|release'
t_WEAK = r'W|weak'
t_GLOBALLY = r'G|globally|always'
t_NEXT = r'X|next'
t_EVENTUALLY = r'F|eventually'
t_FALSE = r'\#|false'
t_TRUE = r'\@|true'
t_PROP = r'[a-z]+'
t_DEMONIC = r'<J[1-9]\d*>'

# Token error handling
def t_error(t):
    t.lexer.skip(1)
    
class DemonicValueError(Exception):
    pass
    
# Grammar
def p_expression_binary(p):
    '''expression : expression AND expression
                  | expression OR expression
                  | expression IMPLIES expression'''
    p[0] = (p[2], p[1], p[3])

def p_expression_ternary(p):
    '''expression : DEMONIC expression UNTIL expression
                  | DEMONIC expression WEAK expression
                  | DEMONIC expression RELEASE expression'''
    demonic_cost = re.findall(r'\d+', p[1])[0]
    try:
        int(demonic_cost)
    except ValueError:
        raise DemonicValueError("Provided cost ({demonic_cost}) is not an int.")
    p[0] = (p[1] + p[3], p[2], p[4])


def p_expression_unary(p):
    '''expression : DEMONIC GLOBALLY expression
                  | DEMONIC NEXT expression
                  | DEMONIC EVENTUALLY expression'''
    demonic_cost = re.findall(r'\d+', p[1])[0]
    try:
        int(demonic_cost)
    except ValueError:
        raise DemonicValueError("Provided cost ({demonic_cost}) is not an int.")
    p[0] = (p[1] + p[2], p[3])


def p_expression_not(p):
    '''expression : NOT expression'''
    p[0] = (p[1], p[2])


def p_expression_group(p):
    '''expression : LPAREN expression RPAREN'''
    p[0] = p[2]

def p_expression_boolean(p):
    '''expression : FALSE 
                  | TRUE'''
    p[0] = p[1]


def p_expression_prop(p):
    '''expression : PROP'''
    p[0] = p[1]


def p_error(p):
    pass


# lexer and parser
lexer = lex.lex()
parser = yacc.yacc()


def get_lexer():
    return lexer


# given an ATL formula as input
# returns a tuple representing the formula divided into subformulas.
def do_parsing(formula):
    try:
        result = parser.parse(formula)
        return result
    except SyntaxError as e:  # if parser fails
        return None
    except DemonicValueError: # invalid cost
        return None


# checks whether the input operator corresponds to a given operator defined in the grammar
def verify(token_name, string):
    lexer.input(string)
    for token in lexer:
        if token.type == token_name and token.value in string:
            return True
    return False



# allow to see if all the eements of a list a are in a list b
def into(lista, listb):
    for i in range(len(lista)):
        temp = False
        for j in range(len(listb)):
            if lista[i] == listb[j]:
                temp = True
        if temp == False:
            return False
    return True 



### Import packages
import os
from pathlib import Path
import base64
import numpy as np

import graphviz

from PIL import Image, ImageDraw, ImageFont
from htbuilder import HtmlElement, div, ul, li, br, hr, a, p, img, styles, classes, fonts
from htbuilder.units import percent, px
from htbuilder.funcs import rgba, rgb

import streamlit as st

from class_game import *
from back_end_CS import *
from front_end_CS import *

global page
st.set_page_config(
    page_title="Agent Theory", layout="wide", page_icon="images/flask.png"
)

### methods used within streamlit - to organize later on


def img_to_bytes(img_path):
    img_bytes = Path(img_path).read_bytes()
    encoded = base64.b64encode(img_bytes).decode()
    return encoded

###Cases Studies

def display_page(page):
  if page==2:
    display_MCMAS()
  elif page == 0 or page==3 or page==4 or page==5:
    # ms_steps = st.selectbox(' ', ['01 - Create MAS', '02 - Upload File'])
    display_MS(page)
  else:
    nlp_steps = st.selectbox(' ', ['01 - Initialization',
                                        '02 - Strategy Example', '03 - ICGS','04 - Dining Cryptographers',
                                        '05 - Buchi  Automaton', '06 - Upload Configuration'])
    display_case(nlp_steps)

### Main Front

def main():
    def _max_width_():
        max_width_str = f"max-width: 1000px;"
        st.markdown(
            f"""
        <style>
        .reportview-container .main .block-container{{
            {max_width_str}
        }}
        </style>
        """,
            unsafe_allow_html=True,
        )


    # Hide the Streamlit header and footer
    def hide_header_footer():
        hide_streamlit_style = """
                    <style>
                    footer {visibility: hidden;}
                    </style>
                    """
        st.markdown(hide_streamlit_style, unsafe_allow_html=True)

    # increases the width of the text and tables/figures
    _max_width_()

    # hide the footer
    hide_header_footer()

    st.markdown("# VITAMIN")
    st.subheader(
        """
        VerIficaTion of A MultI ageNt system
        """
    )
    st.markdown("     ")




    selected_indices = []


    index_review = 0

    # st.markdown(
    #     """
    #     [<img src='data:image/png;base64,{}' class='img-fluid' width=25 height=25>](https://github.com/hi-paris/agent-theory) <small> vitamin 0.0.1 | September 2022</small>""".format(
    #         img_to_bytes("./images/github.png")
    #     ),
    #     unsafe_allow_html=True,
    # )

    #Useful Session State
    if 'info_model' not in st.session_state:
      st.session_state.info_model=[]
    if 'mat_transi' not in st.session_state:
      st.session_state.mat_transi=[]
    if 'cmpt_model' not in st.session_state:
      st.session_state.cmpt_model=0
    if 'page' not in st.session_state:
      st.session_state.page=3
    st.sidebar.header("Dashboard")
    st.sidebar.markdown("---")
    if st.sidebar.button("0 - Guide"):
       st.session_state.page=0
    st.sidebar.markdown("---")
    st.sidebar.markdown("Formal Verification of MAS")
    st.sidebar.header("Who are you?")
    if st.sidebar.button('1 - Non-Expert User'):
      st.session_state.cmpt_model=0
      st.session_state.info_model=[]
      st.session_state.mat_transi=[]
      st.session_state.costs=[]
      st.session_state.page=3
    if st.sidebar.button('2 - Expert User'):
      st.session_state.cmpt_model=0
      st.session_state.info_model=[]
      st.session_state.mat_transi=[]
      st.session_state.costs=[]
      st.session_state.page=4
    # st.sidebar.markdown("---")
    # st.sidebar.markdown("Formal Verification of Attack Graphs")
    # if st.sidebar.button('3 - Expert User'):
    #   st.session_state.page=5
    #   st.session_state.cmpt_model=-1
    #   st.session_state.info_model_test=[]

    display_page(st.session_state.page)







if __name__=='__main__':
    main()

st.markdown(" ")
st.markdown("### ** 👨🏼‍💻 Developers and Researchers: **")
st.image(['images/vadim.jpg', 'images/angelo.jpg', 'images/giulia.jpg', 'images/marco.jpg'], width=110, caption=["Vadim Malvone", "Angelo Ferrando", "Giulia Luongo", "Marco Aruta"])

# st.image(images, width=250)
# st.write('    ')
# st.markdown('### Contributors:')

st.markdown('### In collaboration with Hi!Paris')
images = Image.open('./images/hi-paris.png')
PA=Image.open('./images/PA.jpg')
Pierre=Image.open('./images/Pierre.jpg')
GAE=Image.open('./images/gaetan.png')
st.image([PA,GAE,Pierre],width=110)


st.markdown(f"####  Link to Project Website [here]({'https://github.com/VadimMalvone/VITAMIN'}) 🚀 ")



def image(src_as_string, **style):
    return img(src=src_as_string, style=styles(**style))


def link(link, text, **style):
    return a(_href=link, _target="_blank", style=styles(**style))(text)


def layout(*args):

    style = """
    <style>
      # MainMenu {visibility: hidden;}
      footer {visibility: hidden;background - color: white}
     .stApp { bottom: 80px; }
    </style>
    """
    style_div = styles(
        position="fixed",
        left=0,
        bottom=0,
        margin=px(0, 0, 0, 0),
        width=percent(100),
        color="black",
        text_align="center",
        height="auto",
        opacity=1,

    )

    style_hr = styles(
        display="block",
        margin=px(8, 8, "auto", "auto"),
        border_style="inset",
        border_width=px(2)
    )

    body = p()
    foot = div(
        style=style_div
    )(
        hr(
            style=style_hr
        ),
        body
    )

    st.markdown(style, unsafe_allow_html=True)

    for arg in args:
        if isinstance(arg, str):
            body(arg)

        elif isinstance(arg, HtmlElement):
            body(arg)

    st.markdown(str(foot), unsafe_allow_html=True)

def footer2():
    myargs = []
    layout(*myargs)


if __name__ == "__main__":
    footer2()



from itertools import combinations
from vitamin_model_checker.model_checker_interface.explicit.CapATL.pre import *
from vitamin_model_checker.logics.CapATL import *
from vitamin_model_checker.model_checker_interface.explicit.CapATL.classes import *
from vitamin_model_checker.models.capCGS import *

from vitamin_model_checker.models.capCGS.capCGS import *
import itertools
from itertools import combinations, permutations
from vitamin_model_checker.model_checker_interface.explicit.CapATL.classes import *

##########################################################
# donne toutes le combinaisons de sous listes 
# i.e. [[1, 5], [4], [2,3]] donne [(1, 4, 2), (1, 4, 3), (5, 4, 2), (5, 4, 3)]
def trouver_combinaisons(listes):
    combinaisons = list(itertools.product(*listes))
    return combinaisons


# returns the states where the proposition holds
def get_states_prop_holds(prop):
    states = set()
    prop_matrix = cCGS.get_matrix_proposition()

    index = get_atom_index(prop)
    if index is None:
        return None
    for state, source in enumerate(prop_matrix):
        if source[int(index)] == 1:
            states.add(state)
    return states


#returns the atomic propositions held in the state 
def get_prop_held_in_state(state):
    props = set()
    prop_matrix = cCGS.get_matrix_proposition()

    index = get_index_by_state_name(state)
    if index is None : 
        return None
    state_atoms=prop_matrix[index] # row of the state
    for prop, source in enumerate(state_atoms):
        if source == 1 :
            props.add(prop)
    return props 


# returns the index of the given atom, in the array of atomic propositions
def get_atom_index(element):
    array = cCGS.get_atomic_prop()
    try:
        index = np.where(array == element)[0][0]
        return index
    except IndexError:
        print("Element not found in array.")
        return None

# returns the index given a state name 
def get_index_by_state_name(state):
    if len(state) ==1 :
        state = state[0]
    else :
        interm = ''
        for elem in state : 
            interm = interm + elem
        state = str(interm)

    state_list = cCGS.get_states()

    for count, st in enumerate(state_list) :
        if state == st :
            return count

# returns the state, given an index
def get_state_name_by_index(index):
    states = cCGS.get_states()
    return states[index]


# converts action_string into a list
def build_list(action_string):
    ris = ''
    if action_string == '*':
        for i in range(0, cCGS.get_number_of_agents()):
            ris += '*'
        action_string = ris
    action_list = action_string.split(',')
    return action_list


# returns a set of agents given a coalition ("1,3,6") -> {'6','1','3'}
def get_agents_from_coalition(coalition):
    agents = coalition.split(",")
    return set(agents)


# sort and remove 0 from agents
def format_agents(agents):
    agents = sorted(agents)
    if 0 in agents:
        agents.remove(0)
    agents = {int(x) - 1 for x in agents}
    return agents


# returns possible capacities given an action A or B etc (not a vector)
def get_capacities_from_action(action):
    possible_capacities = []
    capacities = cCGS.get_action_capacities() 
    for tupl in capacities:
        if action in tupl: 
            possible_capacities.append(tupl[0])
    return possible_capacities


def get_capacities_from_action2(action, agent) :
    result = []
    cap_ag = cCGS.get_capacities_assignment()[int(agent)-1][1:]
    ens = cCGS.get_action_capacities()
    if action == '*' :
        result = cap_ag
    for j in ens :
        if j[0] in cap_ag :
            if action in j :
                result.append(j[0])
    return result


def get_capacities_from_actionvector(action, agents):
    result1 = []
    for ag in agents :
        interm = []
        act = action[int(ag)-1]
        interm = get_capacities_from_action2(act, ag)
        result1.append(interm)
    result = trouver_combinaisons(result1)
    return result


# returns the set X_agt_cap
def X_agt_cap():
    cap_assgn = cCGS.get_capacities_assignment()
    interm = []
    result = []
    for elem in cap_assgn :
        interm.append(elem[1:])
    interm = trouver_combinaisons(interm)
    for value in interm :
        result.append(list(value))

    return result

def X_agt_cap2() : 
    result = set()
    for elem in X_agt_cap():
        result.add(tuple(elem))
    return result


def generate_subsets(liste):
    subsets = []
    for r in range(1, len(liste) + 1):
        subsets.extend(tuple(tuple(subset) if len(subset) > 1 else subset[0] for subset in combinations(liste, r)))
    return subsets

def possible_knowledge() :
    result_f = []
    capacity_ass = X_agt_cap2()
    result_f = generate_subsets(capacity_ass)
    return result_f


def generer_sous_ensembles_ord_rep(ensemble, p):
    n = len(ensemble)
    resultats = []

    def generer_sous_ensemble_ord_rep_recursion(sous_ensemble):
        if len(sous_ensemble) == p:
            resultats.append(sous_ensemble)
        elif len(sous_ensemble) < p:
            for i in range(n):
                generer_sous_ensemble_ord_rep_recursion(sous_ensemble + [ensemble[i]])

    generer_sous_ensemble_ord_rep_recursion([])
    return resultats


def intersection_listes(listes):
    if not listes or len(listes) < 2:
        return []
    # Convertir la première liste en un ensemble de tuples
    ensemble_intersection = set(map(tuple, listes[0]))
    # Trouver l'intersection avec les autres listes
    for liste in listes[1:]:
        ensemble_intersection = ensemble_intersection.intersection(map(tuple, liste))
    # Convertir l'ensemble d'intersection en une liste de listes
    intersection = list(map(list, ensemble_intersection))
    
    return intersection


# returns coalition's actions from a given state 
def get_coalition_action_from_q(coalition, state_q) :

    agents = get_agents_from_coalition(coalition)
    agents = format_agents(agents)

    graph = cCGS.get_graph()
    if get_index_by_state_name(state_q) is None : 
        return None
    possible_actions = graph[get_index_by_state_name(state_q)]
    result = []
    string = ""
    index = 0
    for act in possible_actions :
        if type(act) == str :
            for ag in agents :
                index = int(ag)
                string = string + act[index]
            result.append(string)
            string = ""
    return result   


def pointed_knowledge_set() : 
    P_cap = possible_knowledge()
    if [] in P_cap : 
        P_cap.remove([])
    for j in P_cap : 
        j = tuple(j)
    Pcap = generer_sous_ensembles_ord_rep(P_cap, cCGS.get_number_of_agents())
    states = cCGS.get_states()
    result = []
    agents = tuple(i for i in range(1,cCGS.get_number_of_agents()+1))

    for st in states : 
        for value in Pcap :
            if intersection_listes(value) != [] :
                obj = p_knowledge(tuple(st), tuple(value), agents)
                result.append(obj)  
    return result


def Omega_Y(coal_Y) : #coal_Y string "1,3"
    result = []
    coalition = coal_Y
    pointed_knowl = pointed_knowledge_set()
    agents_tot = tuple(i for i in range (1, cCGS.get_number_of_agents()+1))
    
    for elem in pointed_knowl :
        poss_action = get_coalition_action_from_q(coal_Y, getattr(elem, 'state'))
        if poss_action is not None :
            for act in poss_action :
                obj = p_knowledge_for_Y(0,[],0,[],[])
                obj.coalition = coalition
                obj.agents = agents_tot
                obj.state = getattr(elem, 'state') 
                obj.knowledge = getattr(elem, 'knowledge')
                obj.action = act
                result.append(obj)

    return result


# returns a list of possible actions for elem = [q, delta1...deltak] so that elem + action is in W
def action_in_W(elem, W) :
    action = []

    for value in W :

        if getattr(value, 'knowledge') == getattr(elem, 'knowledge') and getattr(value, 'state') == getattr(elem, 'state'):
            action.append(getattr(value, 'action'))
    return action


#function pi_theta needed in model checking
def pi_theta(W) :
    theta = pointed_knowledge_set()
    result = []
    for elem in theta :
        if len(action_in_W(elem, W)) :
            result.append(elem)
    return result 


# returns the actions that corresponds to the given capacity
def get_actions_from_capacity(capacity):
    result = 0
    corresp = cCGS.get_action_capacities()
    for value in corresp : 
        if value[0]==capacity :
            result = value[1:]
    return result


# returns coalition's actions
def get_coalition_action(coalition):
    agents = get_agents_from_coalition(coalition)
    
    cap_ag = cCGS.get_capacities_assignment()
    result = []
    act = []
    for ag in agents : 
        
        capacities = cap_ag[int(ag)-1][1:]
        for capac in capacities :
            act += get_actions_from_capacity(capac)
        result.append(act)   
        act = []
    return trouver_combinaisons(result)


# retruns the non null actions to reach q 
def get_actions_to_reach_q(state_q) :
    result = []
    graph = cCGS.get_graph()
    nb_states = len(get_states())
    nb_agents = cCGS.get_number_of_agents()
    index = get_index_by_state_name(state_q)

    for i in range(0, nb_states) :
        act = graph[i][index]
        if act != 0 :
            if len(act)>nb_agents :
                interm = act.split(',')
                for elem in interm :
                    result.append(elem)
            else : 
                result.append(act)
    return result


def intersection_n_listes(listes):

    ensembles = [set(lst) for lst in listes]
    intersection = set.intersection(*ensembles)
    return intersection


def possible_capacity(liste, taille ):
    sub_set = generer_sous_ensembles_ord_rep(liste, taille)
    result = []

    for elem in sub_set :
        if len(intersection_n_listes(elem)) != 0 :
            result.append(elem)
    return result


def pi_omega_Y(W, coalition_Y) :
    omega_Y = Omega_Y(coalition_Y)
    result = []
    dict_W = {}
    for obj in W : 
        key = (obj.state, obj.knowledge)
        dict_W[key] = obj
    for value in omega_Y :
        key = (value.state, value.knowledge)
        if key in dict_W :
            result.append(value)
    return result


# returns les actions de s1 à s2 que agent ne peut distinguer 
# donne les actions entre deux états ou agent joue le même rôle (fait la même action) que dans action
def indistinguishable_action(state1, state2, action, agent):
    action_indisting = []
    graph = cCGS.get_graph()
    index1 = get_index_by_state_name(state1)
    index2 = get_index_by_state_name(state2)
    if index1 == None or index2 == None :
        print("index faux")
        return None

    
    actions_s1tos2 = graph[index1][index2]
    if str(action) not in str(actions_s1tos2):
        return None 
    # si l'action précisée ne peut faire la transition de state1 à state2
    # if agent not in actions_s1tos2 :
    #     return None
    # si l'agent ne peut effectuer l'action précisée pour passer de state1 à state 2
    liste_actions1tos2 = build_list(actions_s1tos2)
    for act in liste_actions1tos2 : 
        if act[int(agent)-1]==action[int(agent)-1] :
            action_indisting.append(act)
        ## actions qui permettent d'aller de s1 à s2 et où action de l'agent a est la même que action
    return action_indisting #retourne aussi l'action entrée


def function_F_for_succ(q1, q2, alpha, agent_a):
    possible_action = indistinguishable_action(q1, q2, alpha, agent_a)
    agents = list(range(1,cCGS.get_number_of_agents()+1))

    if possible_action == None :
        return None

    possible_set_capacities = []
    for act in possible_action :
        possible_cap = []
        possible_cap = get_capacities_from_actionvector(act, agents)
            
        possible_set_capacities += possible_cap
    
    return possible_set_capacities


def succ(p_knowledge_for_Y):
    S = []

    coalition = getattr(p_knowledge_for_Y, 'coalition')
    agents_coal = get_agents_from_coalition(coalition)
    action = getattr(p_knowledge_for_Y, 'action')
    state = getattr(p_knowledge_for_Y, 'state')
    set_capacity = getattr(p_knowledge_for_Y, 'knowledge')
    graph = cCGS.get_graph()

    act_st = [] # actions of the coalition that lead to an accessible state
    accessible_state = [] #state accessible from q
    act_coalition = get_coalition_action(coalition)
    index = get_index_by_state_name(state)
    
    if index is None :
        return None 
    
    # accessible state list build by checking in the row of state if there is an action of the coalition
    for count, value in enumerate(graph[index]):
        if value != 0 :
            accessible_state.append(count)
            act_st.append(value)

    act_interm = []
    action_acceptable = []
    action_acceptable_final = []
    k = 0
    non_acc_state = []
    # need to check if these actions match with beta for all the coalition
    for supp, act in enumerate(act_st) : 
        action_acceptable = []
        act_interm = build_list(act) # list of possible actions from q to act

        for count, elem in enumerate(act_interm) :
            for count2, value2 in enumerate(list(format_agents(set(agents_coal)))): 
                if str(elem[value2]) != str(action[count2]):
                    k +=1 
            
            if k==0 : # each action corresponds to beta 
                action_acceptable.append(elem)
            k = 0
        
        if len(action_acceptable) == 0 : # no action to reach the state corresponds to beta 
            non_acc_state.append(accessible_state[supp])
        else :
            action_acceptable_final.append(action_acceptable)
        act_interm = []

    accessible_state = [ x for x in accessible_state if x not in non_acc_state]
    capacity_interm = [] # delta,a'
    set_capacity2 = [] # set of delta,a'
    intersection = set_capacity
    z = []
    agents_tot = list(range(1,cCGS.get_number_of_agents()+1)) #[1,2,...]
    for count2bis, q in enumerate(accessible_state) :   
        for alpha in action_acceptable_final[count2bis] :
            set_capacity2 = []
            # g, d = 0
            for a in agents_tot : 
                z = []
                if function_F_for_succ(state, get_state_name_by_index(q), alpha, a) == None :
                    capacity_interm = []
                
                else :
                    if type(set_capacity[a-1][0]) != tuple :
                        g = {set_capacity[a-1]}
                    else : 
                        g = set(set_capacity[a-1])
                    if len(function_F_for_succ(state, get_state_name_by_index(q), alpha, a)) == 1 : 
                        d = {function_F_for_succ(state, get_state_name_by_index(q), alpha, a)[0]}
                    else : 
                        d = set(function_F_for_succ(state, get_state_name_by_index(q), alpha, a))
                    capacity_interm = g & d 
                for u in capacity_interm : 
                    z.append(u)
                set_capacity2.append(z)
            intersection = intersection_n_listes(set_capacity2)
            set_capacity2 = tuple((sublist[0]) if len(sublist) == 1 else tuple(sublist) for sublist in set_capacity2 )
            obj = p_knowledge(get_state_name_by_index(q), set_capacity2, agents_tot )
            if len(intersection) != 0 and obj.not_in(S) :
                S.append(obj)
    return S


def point_know_in_set(point_know_subset, Set) :
    result = []
    for elem in Set : 
        if elem.not_in(point_know_subset) :
            result.append(elem)
    return result


# returns the possible actions for agent in state
def protocol_function(agent, state) :
    return get_coalition_action_from_q(str(agent), state)


# return if there is some possible set of action to do in st
def final_state(st) :
    graph = cCGS.get_graph()
    if get_index_by_state_name(st) is not None :
        for j in graph[get_index_by_state_name(st)] :
            if j != 0 :
                return True
    return False 

# returns true if the model is correct otherwise false
def validity_model() :
    # 1st condition 
    for ag in range(1, cCGS.get_number_of_agents()) :
        for st in cCGS.get_states() :
            act = []
            act_graph = protocol_function(ag, st)
            cap = cCGS.get_capacities_assignment()[ag-1]
            for elem in cap[1:] :
                for j in get_actions_from_capacity(elem) :
                    act.append(j)
            for x in act_graph : 
                if x not in act :
                    return False
    
    # 2nd condition 
    graph = cCGS.get_graph()
    for ag in range(1, cCGS.get_number_of_agents()) :
        for st in cCGS.get_states() :
            if final_state(st) is not False :
                for c in cCGS.get_capacities_assignment()[ag-1][1:] :
                    intersec = [l for l in protocol_function(ag, st) if l in get_actions_from_capacity(c)]
                    if len(intersec) == 0 :
                        print(st, c, ag)
                        return False

    return True


def verifier_chiffres_et_lettres(liste):
    a_des_chiffres = False
    a_des_lettres = False
    if liste == None :
        return None
    for element in liste:
        if isinstance(element, int) or element.isdigit():
            a_des_chiffres = True
        elif isinstance(element, str) and element.isalpha():
            a_des_lettres = True

    return a_des_chiffres and a_des_lettres


# returns the complement of point_know_subset in set, i.e. all the elements of set that are not in subset
def point_know_in_set(point_know_subset, Set) :
    result = []
    for elem in Set : 
        if elem.not_in(point_know_subset) :
            result.append(elem)
    return result


# returns the intersection of two sets of pointed knowledge 
def intersection_set_pointkno(set1, set2) :
    result = []
    for elem in set1 :
        if elem.not_in(set2) == False :
            result.append(elem)
    return result


# retruns true if set1 is in or equal to set2 
def set_in(set1, set2) : 
    size1 = len(set1)
    size2 = len(set2)
    if size1 > size2 :
        return False 
    for elem in set1 :
        if elem.not_in(set2) == True :
            return False
    return True 


# returns given a tpl 36c1 ag=36
def get_ag_capag(tpl) :
    result = ""
    for elem in tpl : 
        if (elem.isdigit() == False) :
            return result 
        result = result + elem
    return int(result)


# returns the subset of ensemble = pointed knowledge such as the actions corresponds to the cap of agent
def select(ensemble, agent, cap) : 
    ris2 = []
    list_ag = agent.split(',')
    cap = cap.split(',')
    for i in ensemble : 
        k=0
        act = getattr(i, 'action')
        for count, value in enumerate(act) :
            capacity = capacities_from_agact(list_ag[count], value)
            if cap[count] in capacity : 
                k +=1
        if k == len(act) :
            ris2.append(i)
    return ris2


def build_tree(tpl):
    if isinstance(tpl, tuple):
        root = Node_PK(tpl[0])
        if len(tpl) > 1:
            left_child = build_tree(tpl[1])
            if left_child is None:
                return None
            root.left = left_child
            if len(tpl) > 2:
                right_child = build_tree(tpl[2])
                if right_child is None:
                    return None
                root.right = right_child
    else:
        if verifier_chiffres_et_lettres(tpl) :
            phi = []
            capacity = set()
            ag = int(get_ag_capag(tpl)) - 1
            cap_a = tpl[len(str(ag)):]
            X_ag_cap = X_agt_cap()
            for cap in X_ag_cap :
                if cap[ag] == cap_a : 
                    phi.append(cap)
            for elem in phi :
                capacity.add(tuple(elem))
            if len(capacity) == 0 : 
                return None 
            root = Node_PK(capacity)
        else :
            phi = []
            pointed_know = set()
            Theta = pointed_knowledge_set()
            states = get_states_prop_holds(str(tpl))
            if states is None : 
                return None 
            for theta in Theta :
                if get_index_by_state_name(getattr(theta, 'state')) != None :
                    if get_index_by_state_name(getattr(theta, 'state')) in states :
                        phi.append(theta)
            if phi is [] :
                return None 
            else : 
                for elem in phi :
                    pointed_know.add(elem)
                root = Node_PK(pointed_know)
    return root


# returns the coalition string given <1,2,3>jjdhdj
def get_coalition(string) :
    result = ""
    for elem in string[1:] : 
        if elem == '>' :
            return result 
        result = result + elem
    return result


# returns all the numbers of the given string 
def get_agents(string) :
    result =""
    for elem in string :
        if elem.isdigit() :
            result = result + elem
    return result 


# returns [1,2,3], [cap,cop,get] given :  1,2,3 is cap,cop,get
def extract_values_and_words(chaine):
    parties = chaine.split(' is ')
    valeurs_f = parties[0]
    mots_f = parties[1]
    return valeurs_f, mots_f


def capacities_from_agact(ag, act) : 
    capacities_agent = cCGS.get_capacities_assignment()[int(ag)-1][1:]
    if act == '*' :
        return capacities_agent
    cap = get_capacities_from_action(act)
    result = [ x for x in capacities_agent if x in cap]
    return result 


def is_sublist_present(liste_principale, sous_liste):
    set1 = set(liste_principale)
    if type(list(sous_liste)[0]) != tuple :
        set2 = {sous_liste}
    else :
        set2 = set(sous_liste)
    return set2.issubset(set1)
    

def solve_tree(node):
    if node.left is not None:
        solve_tree(node.left)
    if node.right is not None:
        solve_tree(node.right)
    if node.right is None:   # UNARY OPERATORS: not, globally, next, eventually

        if len(str(node.value)) > 32 :
            return node.value

        if verify('NOT', str(node.value)): 
             # e.g. ¬φ
            states = node.left.value # set 
            if type(next(iter(states))) == tuple : 
                ris = []
                Xagtcap = X_agt_cap() # list of lists
                for elem in Xagtcap : 
                    if tuple(elem) not in states : 
                        ris.append(tuple(elem))
                node.value = set(ris)

            if type(next(iter(states))) == p_knowledge :
                pointed_know = set(pointed_knowledge_set())
                ris = pointed_know - states
                node.value = set(ris)
        

        elif verify('COALITION', str(node.value)) and verify('NEXT', str(node.value)):  # e.g. <1>Xφ
            agent = str(get_coalition(node.value))
            states = list(node.left.value)
            ris = pi_omega_Y(states, agent)
            ris2 = []
            list_ag = agent.split(',')
            ris = pre(ris, agent)
            ris = pi_theta(ris)
            node.value = set(ris)


        elif verify('KCAP', str(node.value)) :
            ag = get_agents(node.value)
            ag = int(ag) -1 
            result = []
            Theta = pointed_knowledge_set()
            know = 0
            ens = set()
            for elem in node.left.value :
                ens.add(tuple(elem))
            for elem in Theta :
                know = tuple(getattr(elem, 'knowledge')[ag])
                if is_sublist_present(ens, know) :
                    result.append(elem)
            node.value = set(result)

    
    

    if node.left is not None and node.right is not None:  # BINARY OPERATORS: or, and, until, implies
        
        if verify('COALITION', str(node.value)) and verify('RELEASE', str(node.value)) :
            agent = str(get_coalition(node.value))
            
            W1 = Omega_Y(agent)
            W2 = pi_omega_Y(node.right.value, agent)
            cst = pi_omega_Y(node.left.value, agent)
            
            while set_in(W1, W2) == False :
                W1 = intersection_set_pointkno(W1, W2)
                W2 = pre(W1, agent) + cst
            ris = set(pi_theta(W1))
            node.value = ris

        elif verify('COALITION', str(node.value)) and verify('UNTIL', str(node.value)):  # e.g. <1>φUθ
            agent = str(get_coalition(node.value))
            W1 = []
            W2 = pi_omega_Y(node.right.value, agent)
            cst = pi_omega_Y(node.left.value, agent) #a
            k = 0
            while set_in(W2, W1) == False :
                W1 = W1 + W2 
                k += 1
                W2 = intersection_set_pointkno(pre(W1, agent), cst)
            ris = set(pi_theta(W1))
            node.value = ris

        
        elif verify('AND', str(node.value)):  # e.g. φ ^ θ
            states1 = node.left.value
            
            states2 = node.right.value
            ris = set()
            if type(next(iter(states1))) == p_knowledge : 
                ris = states1 & states2
                node.value = ris

                # ris = states1.intersection(states2)
                node.value = ris
            
            if type(next(iter(states1))) == tuple : 
                ris = states1.intersection(states2)
                
    return node.value


def convert_state_set(state_set):
    states = set()
    for elem in state_set:
        position = get_index_by_state_name(elem)
        states.add(int(position))
    return states


# converts a string into a set
def string_to_set(string):
    if string == 'set()':
        return set()
    set_list = string.strip("{}").split(", ")
    new_string = "{" + ", ".join(set_list) + "}"
    return eval(new_string)


# returns whether the result of model checking is true or false in the initial state
def verify_initial_state(initial_state, string):
    if initial_state in string:
        return True
    return False

def model_checking(formula, filename) :
    global cCGS

    debug = ''
    if not formula.strip():
        result = {'res': 'Error: formula not entered', 'initial_state': ''}
        return result

    # model parsing
    cCGS = capCGS()
    cCGS.read_file(filename)

    # check if the model is acceptable
    if validity_model() is False :
        result = {'res': "Incorrect model", 'initial state': '' }


    # formula parsing
    res_parsing = do_parsing(formula, cCGS.get_number_of_agents())

    if res_parsing is None:
        result = {'res': "Syntax Error", 'initial_state': ''}
        print(result)
        return result

    root = build_tree(res_parsing)
    debug += 'root : ' + str(root) + '\n'
    if root is None:
        result = {'res': "Syntax Error: the atom does not exist", 'initial_state': ''}
        print(result)
        return result

    # model checking
    solve_tree(root)
    state = []
    for elem in root.value : 
        debug += 'elem : ' + str(elem) + '\n'
        # elem.print_p_knowledge()
        k = 0
        for know in getattr(elem, 'knowledge') :
            debug += 'know : ' + str(know) + '\n'
            debug += 'know : ' + str(tuple(X_agt_cap2())) + '\n'
            if know == tuple(X_agt_cap2()) : 
                k +=1        
        debug += 'k : ' + str(k) + '\n'
        if k == cCGS.get_number_of_agents() and getattr(elem, 'state') not in state:
            state.append(getattr(elem, 'state'))
    # print(state)
    # print(root.value)
    # # solution
    # initial_state = get_initial_state()
    # bool_res = verify_initial_state(initial_state, root.value)
    result = {'res': str(state)}

    return result






from xml_data.initialisation import *
from xml_data.set_operation import *

#To suppress the space in a string, for vitamin interpretation
def suppr_space(string):
    res = ""
    for caracter in string:
        if caracter == " ":
            res += "_"
        else:
            res += caracter
    return res
            
            
#Get the lines in a file and put them in an array
def recupere_cond (file_cond):
    fichier = open(file_cond, "r")
    res = [line.rstrip('\n') for line in fichier]
    fichier.close()
    return (res)
    
            

#This program take in entry an xml file describing an hypergraph and a file with the conditions that can be temporarly deactivated and output a cgs structure : states and there transitions
            
def to_cgs(file, file_cond):
    #We get the list of all the conditions that can be temporarly deactevated
    cond_deactiv = recupere_cond(file_cond)
    #We get the data on the hypergraph computed from the xml
    (matrix, attacks, condition) = initialisation(file)
    nb_state_graph = len(matrix)
    nb_condition = len(condition)
    initial_state_flag = [1 for i in range(nb_condition)]
    for i in range(nb_condition):
        #We look in all states : a state without previous condition will be part of the initial state of the cgs
        for j in range(nb_state_graph):
            if matrix[j][condition[i].id]:
                initial_state_flag[i] = 0
    #each state will be composed of : id, conditions, list of successors
    initial_state = (0, [], [])
    for i in range(len(initial_state_flag)):
        if initial_state_flag[i]:
            initial_state[1].append(condition[i].fact)
    initial_state[2].append(0)
    #State array will contains all state that have been fully taken in account (studied each attacks)
    state_array = []
    #transition array will contains the transition swith the full name of : preconditions, postconditions and the attack name
    #It is not useful here but can be of interest for next improvments
    transition_array = []
    #we initiate the list of state to take in account with the initial state
    left_states = [initial_state]
    #this variable will be equal to the number of states added in left_states and so their id (we then start with id  for the initial state)
    nb_states = 0
    #we take in account all the states
    while left_states != []:
        current_state = left_states[0]
        del left_states[0]
        #we take in account all the attacks fo each states
        for attaque in attacks:
            #checking if the attack can be launch from the state
            if  (not into(attaque.postcondition, current_state[1])) and into(attaque.preconditions, current_state[1]):
                new_conditions = []
                for cond in current_state[1]:
                    new_conditions.append(cond)
                new_state = (nb_states + 1, new_conditions, [])
                new_state[1].append(attaque.postcondition[0])
                #we build is transition list from what we know : he doesn't lead to any of the previous states previously added in the left_state flag
                #Otherwise, it would have been already taken into account
                for i in range(nb_states):
                    new_state[2].append(0)
                new_state[2].append(0)
                #We now check if the state already exist
                already_exist = 0
                #we look in the state already taken in account
                for old_state in state_array:
                    #a state is characterisez by it's first list : the conditions that are know by the attacker on it
                    if into(new_state[1], old_state[1]) and into(old_state[1], new_state[1]):
                        #we add the transition in the origin state, to avoid forgeting transitions
                        transition_array.append((suppr_space(attaque.action), current_state, old_state)) 
                        #we then update the transition list of the old state
                        while len(current_state[2]) < old_state[0] + 1:
                            current_state[2].append(0)
                        #we look if the attack can be deactivated
                        deactivable = 0
                        for precond in attaque.preconditions:
                            if precond in cond_deactiv:
                                #if the attack is deactivable, deactivable will be 2
                                deactivable = deactivable + 2 - deactivable
                        current_state[2][old_state[0]] = -1 + deactivable #1 if it's deactivable, -1 otherwise  
                        already_exist += 1
                if already_exist == 0:
                    #we do the same with the states present in left_states
                    for old_state in left_states:
                        if into(new_state[1], old_state[1]) and into(old_state[1], new_state[1]):
                            transition_array.append((suppr_space(attaque.action), current_state, old_state)) 
                            while len(current_state[2]) < old_state[0] + 1:
                                current_state[2].append(0)
                            #we look if the attack can be deactivated
                            deactivable = 0
                            for precond in attaque.preconditions:
                                if precond in cond_deactiv:
                                    #if the attack is deactivable, deactivable will be 2
                                    deactivable = deactivable + 2 - deactivable
                            current_state[2][old_state[0]] = -1 + deactivable #1 if it's deactivable, -1 otherwise
                            already_exist += 1
                #if it pass the test, then it's a new state and we will add it to the list of state which are to be taken in account
                if already_exist == 0:
                    left_states.append(new_state)
                    transition_array.append((suppr_space(attaque.action), current_state, new_state))     
                    while len(current_state[2]) < nb_states + len(left_states) + 1:
                        current_state[2].append(0)
                    #we look if the attack can be deactivated
                    deactivable = 0
                    for precond in attaque.preconditions:
                        if precond in cond_deactiv:
                            #if the attack is deactivable, deactivable will be 2
                            deactivable = deactivable + 2 - deactivable
                    current_state[2][new_state[0]] = -1 + deactivable #1 if it's deactivable, -1 otherwise
                    nb_states += 1               
        state_array.append(current_state)        
 
    return (state_array, transition_array)
    




from setuptools import setup, find_packages

# Function to read the requirements.txt file
def parse_requirements(filename):
    with open(filename) as f:
        return f.read().splitlines()
  
setup( 
    name='vitamin-model-checker', 
    version='1.3', 
    description='The VITAMIN model checker python package', 
    long_description='VITAMIN is an open-source model checker tailored to the verification of Multi-Agent Systems (MAS). MAS descriptions are given by means of labelled transition systems.',
    author='Angelo Ferrando', 
    author_email='angelo.ferrando42@gmail.com', 
    packages=[
        'vitamin_model_checker', 
        'vitamin_model_checker.logics',
        'vitamin_model_checker.logics.ATL', 
        'vitamin_model_checker.logics.CapATL',
        'vitamin_model_checker.logics.CTL',
        'vitamin_model_checker.logics.NatATL',
        'vitamin_model_checker.logics.OATL',
        'vitamin_model_checker.logics.OL',
        'vitamin_model_checker.logics.RABATL',
        'vitamin_model_checker.logics.RBATL',
        'vitamin_model_checker.model_checker_interface',
        'vitamin_model_checker.model_checker_interface.explicit',
        'vitamin_model_checker.model_checker_interface.explicit.ATL',
        'vitamin_model_checker.model_checker_interface.explicit.ATLF',
        'vitamin_model_checker.model_checker_interface.explicit.CapATL',
        'vitamin_model_checker.model_checker_interface.explicit.CTL',
        'vitamin_model_checker.model_checker_interface.explicit.NatATL',
        'vitamin_model_checker.model_checker_interface.explicit.OATL',
        'vitamin_model_checker.model_checker_interface.explicit.OL',
        'vitamin_model_checker.model_checker_interface.explicit.RABATL',
        'vitamin_model_checker.model_checker_interface.explicit.RBATL', 
        'vitamin_model_checker.model_checker_interface.implicit', 
        'vitamin_model_checker.model_checker_interface.abstract', 
        'vitamin_model_checker.models',
        'vitamin_model_checker.models.capCGS',
        'vitamin_model_checker.models.CGS',
        'vitamin_model_checker.models.costCGS'
    ],
    install_requires=parse_requirements('/media/angelo/WorkData/git/VITAMIN-APP/VITAMIN/requirements.txt') 
) 



from xml_data.to_cgs import *

#From an input file (.xml), a file of deactivable condition and a list of goals conditions this function output in the file "file" the txt file that will allow VITAMIN to compute and check the properties we need.
def writting(input, file, file_cond, Objectif_attaque):
    #result transition is of no use, but can serve in net improvments of the program
    (result_state, result_transition) = to_cgs(input, file_cond)
    fichier = open(file+"_Vitamin.txt", "w")
    fichier.write("Transition_With_Costs")
    fichier.close()
    #We write the transitions
    taille = len(result_state)
    #with one line by state
    for i in range(taille):
        transition = ""
        for j in range(taille):
            #If there is no indication then there is no transition
            if j >= len(result_state[i][2]):
                transition += "0 "
                #if the transition is positive then it's deactivable. Otherwise we put it at a maximum : the defenders would need to be omnipotent to deactivate it
            elif result_state[i][2][j] > 0:
                transition += "1 "
            elif result_state[i][2][j] < 0:
                transition += str(taille) + " "
            else:
                transition += "0 "
        fichier = open(file+"_Vitamin.txt", "a")
        fichier.write("\n" + transition)
        fichier.close()         
    
    #We name the different state in the order we write them
    fichier = open(file+"_Vitamin.txt", "a")
    fichier.write("\nName_State")
    fichier.write("\n")
    for i in range(taille):
        fichier.write(str(i) +" ")
    fichier.close()
    
    #there is one atomic proposition : the attacker win
    fichier = open(file+"_Vitamin.txt", "a")
    fichier.write("\nAtomic_propositions")
    fichier.write("\nA_w")
    fichier.close()
   
    #We look with the files of goals for the attacker teh states on which he win
    fichier = open(file+"_Vitamin.txt", "a")
    fichier.write("\nLabelling")
    for i in result_state:
        gagne = 0
        for cond in Objectif_attaque:
            if cond in i[1]:
                gagne = 1
        #if the attacker win then the defender lose.
        fichier.write("\n" + str(1 - gagne) + " " + str(gagne))
    fichier.close()
    
    fichier = open(file+"_Vitamin.txt", "a")
    fichier.write("\nNumber_of_agents")
    fichier.write("\n2")
    fichier.close()
    

writting("./xml_data/test.xml", "./xml_data/test_output", "./xml_data/cond.txt", ['Cond 4'])


from vitamin_model_checker.logics.CTL.parser import *
import re

def natatl_to_ctl(natatl_formula): #transform a NatATL formula into a CTL formula (using "FORALL" path quantifier)
    k_pattern = r'(!?|not)?<\{((?:\d+,)*\d+)\},\s*(\d+)>'
    match = re.search(k_pattern, natatl_formula)
    negation = match.group(1)
    k_value = int(match.group(3))

    ctl_operator = "A"
    ctl_formula = re.sub(k_pattern, ctl_operator, natatl_formula)
    if negation:
        ctl_formula = f"!({ctl_formula})"

    return ctl_formula

def natatl_to_ctl_tree(node):
    pattern = r"<\{((?:\d+,)*\d+)\},\s*(\d+)>"
    if re.search(pattern, node.data):
        node.data = re.sub(pattern, "A", node.data)
    if node.left:
        natatl_to_ctl_tree(node.left)
    if node.right:
        natatl_to_ctl_tree(node.right)


#def get_agents_from_natatl(natatl_formula):
    # Extract k value from NatATL formula
 #   k_pattern = r'(!?|not)?<\{((?:\d+,)*\d+)\},\s*(\d+)>'
  #  match = re.search(k_pattern, natatl_formula)
    # Extract coalition from NatATL formula
   # agents_str = match.group(2)

    # Extract agents from coalition
    #agents = [int(agent) for agent in agents_str.split(',')]

    #return agents


def get_agents_from_natatl(natatl_formula):
    # Extract k value from NatATL formula
    k_pattern = r'(!?|not)?<\{((?:\d+,)*\d+)\},\s*(\d+)>'
    matches = re.findall(k_pattern, natatl_formula)

    # Extract agents from all matches
    agents = set()
    for match in matches:
        agents_str = match[1]
        agents.update(int(agent) for agent in agents_str.split(','))

    return list(agents)

#This function checks if a formula is complex or simple
def negated_formula(input):
    res = do_parsingCTL(input)
    #operators = ["!", "not", "and", "&&", "or", "||", "iff", "implies"]
    operators = ["!", "not"]
    if isinstance(res, tuple):
        return res[0] in operators
    return False

def get_k_value(natatl_formula):
    k_pattern = r'(!?|not)?<\{((?:\d+,)*\d+)\},\s*(\d+)>'
    match = re.search(k_pattern, natatl_formula)
    negation = match.group(1)
    k_value = int(match.group(3))

    return k_value

def replace_formula(formulaCTL, propositions_file):

    # Read the propositions.txt file
    with open(propositions_file, 'r') as file:
        propositions = file.read().split()

    # Replace the CTL formulas with atomic propositions
    temporal_operators = ["AX", "AF"]
    for operator in temporal_operators:
        # Create a pattern that matches the operator followed by any number of alphanumeric characters
        pattern = operator + r'\w*'
        # Find all matches in the formula
        matches = re.findall(pattern, formulaCTL)
        for match in matches:
            # Check if there are still propositions to replace with
            if propositions:
                # Replace the match with the first proposition and remove it from the list
                formulaCTL = formulaCTL.replace(match, propositions.pop(0), 1)
            else:
                print('Not enough propositions to replace all CTL formulas')
                return formulaCTL

    return formulaCTL




from vitamin_model_checker.model_checker_interface.explicit.NatATL.NatATL import model_checking


path = './examples/NatATL/exampleModel1.txt'
# formula = '<{1,3}, 2>Xh'
formula = '!(<{1,2}, 50>X(a and b) and <{1,2}, 50>Xc)'
# formula = '<{1,2}, 1>hUb'

#main function
model_checking(path, formula)






# parsetab.py
# This file is automatically generated. Do not edit.
# pylint: disable=W,C,R
_tabversion = '3.10'

_lr_method = 'LALR'

_lr_signature = 'AND COALITION EVENTUALLY GLOBALLY IMPLIES LPAREN NEXT NOT OR PROP RPAREN UNTILexpression : expression AND expression\n                  | expression OR expression\n                  | expression IMPLIES expressionexpression : COALITION expression UNTIL expressionexpression : COALITION GLOBALLY expression\n                  | COALITION NEXT expression\n                  | COALITION EVENTUALLY expressionexpression : NOT expressionexpression : LPAREN expression RPARENexpression : PROP'
    
_lr_action_items = {'COALITION':([0,2,3,4,6,7,8,10,11,12,18,],[2,2,2,2,2,2,2,2,2,2,2,]),'NOT':([0,2,3,4,6,7,8,10,11,12,18,],[3,3,3,3,3,3,3,3,3,3,3,]),'LPAREN':([0,2,3,4,6,7,8,10,11,12,18,],[4,4,4,4,4,4,4,4,4,4,4,]),'PROP':([0,2,3,4,6,7,8,10,11,12,18,],[5,5,5,5,5,5,5,5,5,5,5,]),'$end':([1,5,13,15,16,17,19,20,21,22,23,],[0,-10,-8,-1,-2,-3,-5,-6,-7,-9,-4,]),'AND':([1,5,9,13,14,15,16,17,19,20,21,22,23,],[6,-10,6,6,6,6,6,6,6,6,6,-9,6,]),'OR':([1,5,9,13,14,15,16,17,19,20,21,22,23,],[7,-10,7,7,7,7,7,7,7,7,7,-9,7,]),'IMPLIES':([1,5,9,13,14,15,16,17,19,20,21,22,23,],[8,-10,8,8,8,8,8,8,8,8,8,-9,8,]),'GLOBALLY':([2,],[10,]),'NEXT':([2,],[11,]),'EVENTUALLY':([2,],[12,]),'UNTIL':([5,9,13,15,16,17,19,20,21,22,23,],[-10,18,-8,-1,-2,-3,-5,-6,-7,-9,-4,]),'RPAREN':([5,13,14,15,16,17,19,20,21,22,23,],[-10,-8,22,-1,-2,-3,-5,-6,-7,-9,-4,]),}

_lr_action = {}
for _k, _v in _lr_action_items.items():
   for _x,_y in zip(_v[0],_v[1]):
      if not _x in _lr_action:  _lr_action[_x] = {}
      _lr_action[_x][_k] = _y
del _lr_action_items

_lr_goto_items = {'expression':([0,2,3,4,6,7,8,10,11,12,18,],[1,9,13,14,15,16,17,19,20,21,23,]),}

_lr_goto = {}
for _k, _v in _lr_goto_items.items():
   for _x, _y in zip(_v[0], _v[1]):
       if not _x in _lr_goto: _lr_goto[_x] = {}
       _lr_goto[_x][_k] = _y
del _lr_goto_items
_lr_productions = [
  ("S' -> expression","S'",1,None,None,None),
  ('expression -> expression AND expression','expression',3,'p_expression_binary','parser.py',43),
  ('expression -> expression OR expression','expression',3,'p_expression_binary','parser.py',44),
  ('expression -> expression IMPLIES expression','expression',3,'p_expression_binary','parser.py',45),
  ('expression -> COALITION expression UNTIL expression','expression',4,'p_expression_ternary','parser.py',50),
  ('expression -> COALITION GLOBALLY expression','expression',3,'p_expression_unary','parser.py',55),
  ('expression -> COALITION NEXT expression','expression',3,'p_expression_unary','parser.py',56),
  ('expression -> COALITION EVENTUALLY expression','expression',3,'p_expression_unary','parser.py',57),
  ('expression -> NOT expression','expression',2,'p_expression_not','parser.py',62),
  ('expression -> LPAREN expression RPAREN','expression',3,'p_expression_group','parser.py',67),
  ('expression -> PROP','expression',1,'p_expression_prop','parser.py',72),
]



import numpy as np

class CGS():
    def init(self):
        self.graph = []
        self.states = []
        self.atomic_propositions = []
        self.matrix_prop = []
        self.initial_state = ''
        self.number_of_agents = ''
        self.actions = []

    def read_file(self, filename):
        with open(filename, 'r') as f:
            lines = f.readlines()

        self.graph = []
        self.states = []
        self.atomic_propositions = []
        self.matrix_prop = []
        self.initial_state = ''
        self.number_of_agents = ''

        current_section = None
        transition_content = ''
        unknown_transition_content = ''
        name_state_content = ''
        atomic_propositions_content = ''
        labelling_content = ''
        rows_graph = []
        rows_prop = []
        

        for line in lines:
            line = line.strip()
            #Section "header"
            if line == 'Transition':
                current_section = 'Transition'
            elif line == 'Unkown_Transition_by':
                current_section = 'Unknown_Transition_by'
            elif line == 'Name_State':
                current_section = 'Name_State'
            elif line == 'Initial_State':
                current_section = 'Initial_State'
            elif line == 'Atomic_propositions':
                current_section = 'Atomic_propositions'
            elif line == 'Labelling':
                current_section = 'Labelling'
            elif line == 'Number_of_agents':
                current_section = 'Number_of_agents'

            #If not header, then read contents based on what section we are in
            
            elif current_section == 'Transition':
                transition_content += line + '\n'
                values = line.strip().split()
                rows_graph.append(values)
            elif current_section == 'Unknown_Transition_by':
                unknown_transition_content += line + '\n'
            elif current_section == 'Name_State':
                name_state_content += line + '\n'
                values = line.strip().split()
                self.states = np.array(values)
            elif current_section == 'Initial_State':
                self.initial_state = line
            elif current_section == 'Atomic_propositions':
                atomic_propositions_content += line + '\n'
                values = line.strip().split()
                self.atomic_propositions = np.array(values)
            elif current_section == 'Labelling':
                labelling_content += line + '\n'
                values = line.strip().split()
                rows_prop.append(values)
            elif current_section == 'Number_of_agents':
                self.number_of_agents = line
            
        actions =[]
        a = 0
        grafo_prov = np.array(rows_graph)
        for row in grafo_prov:
            new_row = []
            for item in row:
                if item == '0':
                    new_row.append(0)
                else:
                    new_row.append(str(item))
                    a = item.split(",")
                    for elem in a :
                        actions.append(elem)
            self.graph.append(new_row)   

        matrix_prop_prov = np.array(rows_prop)
        for row in matrix_prop_prov:
            new_row = []
            for item in row:
                if item == '0':
                    new_row.append(0)
                elif item == '1':
                    new_row.append(1)
                else:
                    new_row.append(str(item))
            self.matrix_prop.append(new_row)

    def read_from_model_object(self, model):
        self.graph = model.transition_matrix
        self.states = np.array(model.state_names)
        self.atomic_propositions = np.array(model.propositions)
        self.matrix_prop = model.labelling_function
        self.initial_state = model.initial_state
        self.number_of_agents = model.number_of_agents
        self.actions = model.actions
        
    def get_graph(self):
        return self.graph

    def get_states(self):
        return self.states

    def get_atomic_prop(self):
        return self.atomic_propositions


    def get_matrix_proposition(self):
        return self.matrix_prop


    def get_initial_state(self):
        return self.initial_state


    def get_number_of_agents(self):
        return int(self.number_of_agents)

    def get_actions(self):
        return self.actions

    def translate_action_and_state_to_key(self, action_string, state):
        return action_string + ";" + state

    def get_actions(self, agents):
        # Convert the graph string to a list of lists
        graph_list = self.graph

        # Create a dictionary to store actions for each agent
        actions_per_agent = {f"agent{agent}": [] for agent in agents}

        for row in graph_list:
            for elem in row:
                if elem != 0 and elem != '*':
                    actions = elem.split(',')
                    for action in actions:
                        for i, agent in enumerate(agents):
                            if action[agent - 1] != 'I':  # idle condition
                                actions_per_agent[f"agent{agent}"].append(action[agent - 1])

        # Remove duplicates from each agent's action list
        for agent_key in actions_per_agent:
            actions_per_agent[agent_key] = list(set(actions_per_agent[agent_key]))

        return actions_per_agent

    #return the number of actions extracted in get_actions()
    def get_number_of_actions (self):

        n = self.get_actions()
        return len(n)

    def write_updated_file(self, input_filename, modified_graph, output_filename):
        if modified_graph is None:
            raise ValueError("modified_graph is None")
        with open(input_filename, 'r') as input_file, open(output_filename, 'w') as output_file:
            current_section = None
            matrix_row = 0
            for line in input_file:
                line = line.strip()

                if line == 'Transition':
                    current_section = 'Transition'
                    output_file.write(line + '\n')
                elif current_section == 'Transition' and matrix_row < len(modified_graph):
                    output_file.write(' '.join([str(elem) for elem in modified_graph[matrix_row]]) + '\n')
                    matrix_row += 1
                elif current_section == 'Transition' and matrix_row == len(modified_graph):
                    current_section = None
                    output_file.write('Unkown_Transition_by' + '\n')
                else:
                    output_file.write(line + '\n')

    #returns the edges of a graph
    def get_edges(self):
        graph = self.get_graph()
        states = self.get_states()
        #duplicate edges (double transactions from "a" to "b") are ignored due to model checking
        edges = []
        for i, row in enumerate(graph):
            for j, element in enumerate(row):
                if element == '*':
                    edges.append((states[i], states[i]))
                elif element != 0:
                    edges.append((states[i], states[j]))
        return edges

    def file_to_string(self, filename):
        with open(filename, 'r') as file:
            data = file.read()
        return data
        
        # returns the index of the given atom, in the array of atomic propositions
    def get_atom_index(self, element):
        array = self.get_atomic_prop()
        try:
            index = np.where(array == element)[0][0]
            return index
        except IndexError:
            print("Element not found in array.")
            return None


    # returns the index, given a state name
    def get_index_by_state_name(self, state):
        state_list = self.get_states()
        index = np.where(state_list == state)[0][0]
        return index


    # returns the state, given an index
    def get_state_name_by_index(self, index):
        states = self.get_states()
        return states[index]


    # converts action_string into a list
    def build_list(self, action_string):
        ris = ''
        if action_string == '*':
            for i in range(0, self.get_number_of_agents()):
                ris += '*'
            action_string = ris
        action_list = action_string.split(',')
        return action_list


    # returns a set of agents given a coalition (e.g. 1,2,3)
    def get_agents_from_coalition(self, coalition):
        agents = coalition.split(",")
        return set(agents)


    # sort and remove 0 from agents
    def format_agents(self, agents):
        agents = sorted(agents)
        if 0 in agents:
            agents.remove(0)
        agents = {int(x) - 1 for x in agents}
        return agents


    # returns coalition's actions
    def get_coalition_action(self, actions, agents):
        coalition_moves = set()
        result = ''
        agents = self.format_agents(agents)
        if len(agents) == 0:
            for i in range(0, self.get_number_of_agents()):
                result += '-'
        else:
            for x in actions:
                div = 2
                # if len(x) > 4:
                #     div = int(len(x) / 4)
                letter_backup = ''
                count = 1
                j = 0
                for _, letter in enumerate(x):
                    if count == div:
                        if j in agents:
                            result += letter if not letter_backup else letter_backup+letter
                        else:
                            result += '-'
                        j += 1
                        count = 1
                        letter_backup = ''
                    else:
                        letter_backup += letter
                        count += 1

                coalition_moves.add(result)
        return coalition_moves


    def get_base_action(self, action, agents):
        return self.get_coalition_action(set([action]), agents).pop()

    # returns all moves except for those of the considered coalition
    def get_opponent_moves(self, actions, agents):
        other_moves = set()
        agents = self.format_agents(agents)
        for x in actions:
            result = ""
            div = 2
            # if len(x) > 4:
            #     div = int(len(x) / 4)
            letter_backup = ''
            count = 1
            j = 0
            for i, letter in enumerate(x):
                if count == div:
                    if j not in agents:
                        result += letter if not letter_backup else letter_backup+letter
                    else:
                        result += '-'
                    j += 1
                    count = 1
                    letter_backup = ''
                else:
                    letter_backup += letter
                    count += 1
            other_moves.add(result)
        return other_moves

    #added NatATL functions below

    def get_label(self, index):
        return f's{index}'

    def create_label_matrix(self, graph):
        label_matrix = []
        for i, row in enumerate(graph):
            label_row = [self.get_label(i) if isinstance(elem, str) and elem != '*' else None for elem in row]
            label_matrix.append(label_row)
        return label_matrix
    
    # Validate transition matrix
    # Use Example
    #matrix = [['III', 0, 0, 0], [0, 'IIZ', 'ADZ,BDZ', 'ACZ,BCI'], ['ACZ,BDZ', 'ICZ', 'III', 'ADZ,BCZ'], [0, 'CIZ', 0, 'III']]
    #n = 3
    #parser(matrix, n)
    def matrixParser(self, n):
        for row in self.graph:
            if all(elem == 0 for elem in row):
                raise ValueError("All row elements are 0")

            char_I_count = [0] * n

            for elem in row:
                if elem == 0:
                    continue

                strings = str(elem).split(',')
                for s in strings:
                    #if len(s) != n:
                    #    raise ValueError(f"string length {s} for element {elem} is not equal to {n}")

                    for i in range(n):
                        if s[i] == 'I':
                            char_I_count[i] += 1

            if any(count == 0 for count in char_I_count):
                raise ValueError("Idle error: There has to be at least one 'I' for each row")